// Copyright 2023 Vincent Jacques

#ifndef LINCS__LEARNING_HPP
#define LINCS__LEARNING_HPP

#include "learning/weights-profiles-breed-mrsort.hpp"
#include "learning/weights-profiles-breed-mrsort/improve-profiles/accuracy-heuristic.hpp"
#include "learning/weights-profiles-breed-mrsort/initialize-profiles/probabilistic-maximal-discrimination-power-per-criterion.hpp"
#include "learning/weights-profiles-breed-mrsort/optimize-weights/glop.hpp"
#include "learning/weights-profiles-breed-mrsort/terminate/at-accuracy.hpp"

#endif  // LINCS__LEARNING_HPP
